@@ -1,14 +1,24 @@
+"""
+Simple file inclusion content: You should probably use the media library
+instead.
+"""
+
+import os
+
 from django.db import models
 from django.template.loader import render_to_string
 from django.utils.translation import ugettext_lazy as _
 
+from feincms import settings
 
 class FileContent(models.Model):
-    # You should probably use `feincms.content.medialibrary.models.MediaFileContent`
-    # instead.
+    # You should probably use
+    # `feincms.content.medialibrary.models.MediaFileContent` instead.
 
     title = models.CharField(max_length=200)
-    file = models.FileField(_('file'), upload_to='filecontent')
+    file = models.FileField(
+        _('file'), max_length=255,
+        upload_to=os.path.join(settings.FEINCMS_UPLOAD_PREFIX, 'filecontent'))
 
     class Meta:
         abstract = True
@@ -19,5 +29,5 @@ class FileContent(models.Model):
         return render_to_string([
             'content/file/%s.html' % self.region,
             'content/file/default.html',
-            ], {'content': self})
+            ], {'content': self}, context_instance=kwargs.get('context'))
 
